#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "generation.h"

// ------------------------------------------------------------------------ //

Grid map ;

template* configs ;
array** matches ;
int n_configs = 2;

int render_distance = 3 ;

int player_x = 4;
int player_y = 4;

int player_cx = 0;
int player_cy = 0;

int zoom = 250 ;
int draw_par = 1;

int __width__ = 1200 ;
int __height__ = 800 ;

imgs digits ;
imgs letters ;

// ------------------------------------------------------------------------ //

void one_debug() {
    matches = malloc(sizeof(array*)*4);

    matches[0] = malloc(sizeof(array));
    matches[1] = malloc(sizeof(array));
    matches[2] = malloc(sizeof(array));
    matches[3] = malloc(sizeof(array));

    matches[0][0].arr = malloc(sizeof(int));
    matches[0][0].arr[0] = 0;
    matches[0][0].len = 1;

    matches[1][0].arr = malloc(sizeof(int));
    matches[1][0].arr[0] = 0;
    matches[1][0].len = 1;

    matches[2][0].arr = malloc(sizeof(int));
    matches[2][0].arr[0] = 0;
    matches[2][0].len = 1;

    matches[3][0].arr = malloc(sizeof(int));
    matches[3][0].arr[0] = 0;
    matches[3][0].len = 1;

    configs = malloc(sizeof(template)); 

    configs[0].lines = malloc(sizeof(uint8_t)*8);
    configs[0].lines[0] = 255;
    configs[0].lines[1] = 129;
    configs[0].lines[2] = 129;
    configs[0].lines[3] = 129;
    configs[0].lines[4] = 129;
    configs[0].lines[5] = 129;
    configs[0].lines[6] = 129;
    configs[0].lines[7] = 255;

    configs[0].id = 0;
}

// ------------------------------------------------------------------------ //

bool unpack_coord(uint8_t* lines, int cx, int cy) {
    return (bool)((lines[cy]/pw(2, cx))%2);
}

bool is_compatible(chunk ch, cardinal dir, int config_id) {
    return arr_mem(matches[(int)dir][ch.chdata.id], config_id);
}

void generate_chunk(int x, int y, int config_id) {
    chunk* new = malloc(sizeof(chunk)) ;
    new->chx = x ;
    new->chy = y ;
    new->draw_id = draw_par - 2;
    new->chdata = configs[config_id];
    //printf("               inserting\n");
    gridInsert(map, x, y, new);
}

void select_config(chunk ch, cardinal dir) {
    //printf("           in\n");
    int i = 1 + (rand()%n_configs);
    int idx = 0 ;
    //printf("init : %d\n", i);
    while(i > 0) {
        //printf("  current : %d | %d\n", i, idx);
        if(is_compatible(ch, dir, idx)) {
            i -= 1;
            if(i != 0) {
                idx = (idx+1)%n_configs;
            };
            //printf("    pass\n");
        } else {
            idx = (idx+1)%n_configs;
            //printf("    smash\n");
        }
    };
    //printf("chosen : %d\n\n", idx);
    if(dir == NORTH) {
        generate_chunk(ch.chx, ch.chy - 1, idx);
    } else if(dir == SOUTH) {
        generate_chunk(ch.chx, ch.chy + 1, idx);
    } else if(dir == EAST) {
        generate_chunk(ch.chx + 1, ch.chy, idx);
    } else {
        generate_chunk(ch.chx - 1, ch.chy, idx);
    };
    //printf("           out\n");
}

bool is_compat_one(int cx, int cy, cardinal dir, int idx) {
    if(!gridMem(map, cx, cy)) {
        return true;
    };
    return is_compatible(*gridGet(map, cx, cy), dir, idx);
}

void generate_chunk_all(int cx, int cy) {
    int i = 1 + (rand()%n_configs);
    int idx = 0 ;
    while(i > 0) {
        if(is_compat_one(cx-1, cy, EAST, idx) && is_compat_one(cx+1, cy, WEST, idx) && is_compat_one(cx, cy-1, SOUTH, idx) && is_compat_one(cx, cy+1, NORTH, idx)) {
            i -= 1;
            if(i != 0) {
                idx = (idx+1)%n_configs;
            }
        } else {
            idx = (idx+1)%n_configs;
        }
    };
    generate_chunk(cx, cy, idx);
}

void initialize(SDL_Renderer* renderer) {
    map = newGrid();
    import_letters(renderer);
    import_digits(renderer);
    generate_chunk(0, 0, rand()%n_configs);
}

void destroy() {
    freeGrid(map);
    free_digits(digits);
    free_digits(letters);
}

void parse_one(FILE* ptr, FILE* ptr2, int index) {
    for(int i = 0; i < 8; i++) {
        for(int j = 0; j < 8; j++) {
            int xres = get_integer_plus_align(ptr, ptr2);
            configs[index].lines[i] *= 2 ;
            configs[index].lines[i] += xres ;
        }
    };
    for(int i = 0; i < 4; i++) {
        int len = count_char_in_line(ptr2, ',');
        matches[i][index].len = len ;
        matches[i][index].arr = malloc(sizeof(int)*len);
        int wasted = fgetc(ptr); // always a '['
        for(int j = 0; j <len; j++) {
            int read = get_integer(ptr);
            matches[i][index].arr[j] = read ;
        };
        terminate_line(ptr);
    };
    
    terminate_line(ptr);
    terminate_line(ptr2);
}

void parse_configs(char* filename, int n_conf) {
    // Initializing everything //
    n_configs = n_conf ;

    matches = malloc(sizeof(array*)*4);
    // 4 cardinal directions

    matches[0] = malloc(sizeof(array)*n_conf);
    matches[1] = malloc(sizeof(array)*n_conf);
    matches[2] = malloc(sizeof(array)*n_conf);
    matches[3] = malloc(sizeof(array)*n_conf);
    // compatibilities, 1 for each template

    configs = malloc(sizeof(template)*n_conf);
    // data for each template

    for(int i = 0; i < n_conf; i++) {
        configs[i].id = i;
        configs[i].lines = malloc(sizeof(uint8_t)*8);
        for(int j = 0; j < 8; j++) {
            configs[i].lines[j] = 0;
        }
    }

    // Now for the fun part //
    FILE* ptr = fopen(filename, "r");
    FILE* ptr2 = fopen(filename, "r");

    for(int k = 0; k < n_conf; k++) {
        parse_one(ptr, ptr2, k);
    }
}